# Composite error term
CET_ADDI = "addi"
CET_MULT = "mult"
CET_Categories = {
    CET_ADDI: "Additive composite error term",
    CET_MULT: "Multiplicative composite error term"
}

# Frontier
FUN_PROD = "prod"
FUN_COST = "cost"
FUN_Categories = {
    FUN_PROD: "Production frontier",
    FUN_COST: "Cost frontier"
}

# Return to scale
RTS_VRS = "vrs"
RTS_CRS = "crs"
RTS_Categories = {
    RTS_VRS: "Variable returns to scale",
    RTS_CRS: "Constant returns to scale"
}

# Orientation
ORIENT_IO = "io"
ORIENT_OO = "oo"
ORIENT_Categories = {
    ORIENT_IO: "Input orientation",
    ORIENT_OO: "Output orientation"
}

# Optimization
OPT_LOCAL = "local"
OPT_DEFAULT = None